#!/usr/bin/env python
# Author: SJSU Students

import sys
import heapq
from math import isnan, sqrt

""" Heuristic functions, computing "distance" to the target.
    Since the target is always 0, 0, we can ignore some operations.
    There's an interesting caveat introduced by the wrapping of the grid.
    Consider the following:
    | A B C | A
    | D E F |
    | G H I |
    If we are looking for the distance from I to the inside A,
    we have to also consider the wraparound distance to A outside
    and take their minimum."""
def euclidean(node, width):
    return min(sqrt(node.row ** 2 + node.col ** 2),
               sqrt(node.row ** 2 + (node.col - width) ** 2))

def manhattan(node, width):
    return min(node.row + node.col,
               node.row + abs(node.col - width))

def bupkis(node, width):
    return min(node.row ** 3 + node.col ** 3,
               node.row ** 3 + abs(node.col - width) ** 3)

class Node:
    """ Class representing a node in the maze.
    It has UDLR neighbours and costs. """
    def __init__(self, char, row, col):
        self.char = char
        self.row = row
        self.col = col
        self.pcost = float('inf')
        self.hcost = float('nan' if char == '=' else 'inf')
        self.up = None
        self.down = None
        self.left = None
        self.right = None

    # The string representation is just the char inside.
    def __str__(self):
        return self.char

    # Less-than. Needed for heap ordering.
    def __lt__(self, other):
        return self.hcost < other.hcost

    def process(self, frontier, explored, solution, heuristic, width):
        explored.add(self)
        # Available actions. Directions don't allow down.
        for neighbr in [self.up, self.left, self.right]:
            if not neighbr or neighbr in explored or isnan(neighbr.hcost):
                continue
            if neighbr not in frontier:
                heapq.heappush(frontier, neighbr)
            pcost = self.pcost + 1
            if pcost >= neighbr.pcost:
                continue
            solution[neighbr] = self
            neighbr.pcost = pcost
            neighbr.hcost = pcost + heuristic(neighbr, width)

class Solution(dict):
    """ Dict wrapper representing the solution to a maze.
    Object of this class is returned from Maze.solve() """
    def __init__(self, maze):
        self.maze = maze
        dict.__init__(self)

    def __repr__(self):
        node = self.maze.maze[0][0]
        width = self.maze.width
        steps = []
        while node:
            steps.insert(0, node)
            node = self.get(node, None)

        def put_nario(step):
            text = str(self.maze)
            cut = step.row * (width + 1) + step.col
            return text[:cut] + '@' + text[cut + 1:]
        return '\n\n'.join(map(put_nario, steps))

class NullSolution(Solution):
    def __init__(self, maze):
        Solution.__init__(self, maze)

    def __repr__(self):
        return 'NO PATH'

class Maze:
    """ Representation of the maze.
    Constructor makes graph maze of the text. Has 'solve' method. """
    def __init__(self, text):
        # Build text maze. Strip whitespace first.
        text_no_whitespace = filter(str.strip, text.split('\n'))
        assert text_no_whitespace, "No maze"

        # Build array of nodes
        self.maze = []
        for row, line in enumerate(text_no_whitespace):
            current = []
            for col, char in enumerate(line):
                newnode = Node(char, row, col)
                current.append(newnode)
                if char == '@':
                    self.start = newnode
            self.maze.append(current)

        assert self.start, "No Nario"
        self.start.char = '.'

        # Connect the nodes to form a cylinder.
        self.height = len(self.maze)
        self.width = len(self.maze[0])
        for i in range(self.height):
            for j in range(self.width):
                here = self.maze[i][j]
                there = j - 1
                here.left = self.maze[i][there]
                self.maze[i][there].right = here
                there = (j + 1) % self.width
                here.right = self.maze[i][there]
                self.maze[i][there].left = here
        # Careful for vertical, as numbers go in reverse,
        # and we don't want to wrap around.
        for i in range(1, self.height - 1):
            for j in range(self.width):
                here = self.maze[i][j]
                there = (i + 1)
                here.down = self.maze[there][j]
                self.maze[there][j].up = here
                there = (i - 1)
                here.up = self.maze[there][j]
                self.maze[there][j].down = here
        # Top and bottom row.
        for j in range(self.width):
            self.maze[0][j].down = self.maze[1][j]
            self.maze[1][j].up = self.maze[0][j]
            self.maze[-1][j].up = self.maze[-2][j]
            self.maze[-2][j].down = self.maze[-1][j]

    def __str__(self):
        return '\n'.join(map(lambda row: ''.join(map(str, row)), self.maze))

    def solve(self, heuristic):
        solution = Solution(self)
        explored = set()
        frontier = [self.start] # A simple array, but we use heappush/pop

        self.start.hcost = heuristic(self.start, self.width)
        self.start.pcost = 0
        while frontier:
            node = heapq.heappop(frontier)
            if not heuristic(node, self.width):
                return solution
            node.process(frontier, explored, solution, heuristic, self.width)

        return NullSolution(self.maze)

if __name__ == '__main__':
    # Read file, parse into lines, strip whitespace, convert to Boolean.
    with open(sys.argv[1], 'r') as f:
        # Choose a heuristic
        heuristics = { 'ma' : manhattan, 'eu' : euclidean }
        choice = heuristics.get(sys.argv[2][:2].lower(), bupkis)

        # Run.
        print Maze(f.read()).solve(choice)
